/*
 * RESOURCES.PRC
 *
 * PWLib application resources file for openphone
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1999 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): Portions Copyright 1999 NEWLINK
 *
 * $Log: resources.prc,v $
 * Revision 1.70  2002/07/04 00:48:43  robertj
 * More H.450.11 call intrusion implementation, thanks Aleksandar Todorovic
 *
 * Revision 1.69  2002/06/22 05:49:12  robertj
 * Added partial implementation for H.450.11 Call Intrusion
 *
 * Revision 1.68  2002/05/28 07:13:59  robertj
 * Split UDP (for RAS) from RTP port bases.
 *
 * Revision 1.67  2002/05/21 02:43:40  robertj
 * Added ability to record incoming media to WAV file.
 *
 * Revision 1.66  2002/05/17 06:00:47  robertj
 * Added provision for setting password on secure gatekeeper access.
 *
 * Revision 1.65  2002/05/03 05:39:14  robertj
 * Added Q.931 Keypad IE mechanism for user indications (DTMF).
 *
 * Revision 1.64  2002/04/12 06:08:26  robertj
 * Added ability to set ports for TCP and RTP IDP.
 *
 * Revision 1.63  2002/01/24 07:42:37  robertj
 * Added ability to turn off H.245 negotiations in SETUP option
 *
 * Revision 1.62  2002/01/17 07:03:30  robertj
 * Added support for RFC2833 embedded DTMF in the RTP stream
 *
 * Revision 1.61  2002/01/15 07:35:00  robertj
 * Cleaned up UI messages to allow for VOIP blaster
 *
 * Revision 1.60  2002/01/15 06:39:54  robertj
 * Added support for VOIP blaster
 *
 * Revision 1.59  2001/12/13 11:20:59  robertj
 * Added access token OID user interface for auto set of ACF token to SETUP pdu.
 *
 * Revision 1.58  2001/05/31 01:30:58  robertj
 * Added code to retrieve a held call.
 *
 * Revision 1.57  2001/05/31 00:39:43  robertj
 * Added option to use FACILITY method for call transfer.
 *
 * Revision 1.56  2001/05/23 10:06:58  robertj
 * Added ability to send fake video, thanks Paonza
 *
 * Revision 1.55  2001/05/09 04:07:55  robertj
 * Added more call end codes for busy and congested.
 *
 * Revision 1.54  2001/05/01 05:25:34  robertj
 * Added command to do H.450.x call hold function.
 *
 * Revision 1.53  2001/04/11 03:01:30  robertj
 * Added H.450.2 (call transfer), thanks a LOT to Graeme Reid & Norwood Systems
 *
 * Revision 1.52  2001/04/02 23:58:24  robertj
 * Added jitter calculation to RTP session.
 * Added trace of statistics.
 *
 * Revision 1.51  2001/03/30 09:37:52  robertj
 * Added selection of video capture device.
 *
 * Revision 1.50  2001/03/16 06:46:21  robertj
 * Added ability to set endpoints desired time to live on a gatekeeper.
 *
 * Revision 1.49  2000/12/01 00:09:53  robertj
 * Rearranged sliders for adjusting silence detect delays.
 *
 * Revision 1.48  2000/11/30 05:16:24  robertj
 * Added ability to adjust the silence detection deadband times.
 *
 * Revision 1.47  2000/11/27 00:42:42  robertj
 * Added detection of tones and hookflash on xJACK to be sent via UserIndication.
 * Added forced test for PSTN line on startup.
 *
 * Revision 1.46  2000/10/13 04:17:20  robertj
 * Added clamping to avoid ever setting incorrect tx frame count.
 *
 * Revision 1.45  2000/09/01 02:13:05  robertj
 * Added ability to select a gatekeeper on LAN via it's identifier name.
 *
 * Revision 1.44  2000/08/31 12:45:32  robertj
 * Added ability to set country code on xJack cards.
 *
 * Revision 1.43  2000/08/01 03:24:49  robertj
 * Changed enumeration of Quicknet devices to use new technique for future Win2k drives.
 *
 * Revision 1.42  2000/07/15 09:48:49  robertj
 * Added message for unregistering from gatekeeper.
 *
 * Revision 1.41  2000/07/13 16:01:33  robertj
 * Split autoStartVideo so can select receive and transmit independently
 * Added frame count to message on audio channel open.
 *
 * Revision 1.40  2000/07/11 19:45:51  robertj
 * Fixed crash if closed speed dial window while handset off hook.
 * Added ability to set the frames per packet in audio codecs.
 *
 * Revision 1.39  2000/06/29 11:00:05  robertj
 * Added user interface for sound buffer depth adjustment.
 *
 * Revision 1.38  2000/06/29 06:20:53  robertj
 * Fixed deadlock when exiting application while in a call.
 *
 * Revision 1.37  2000/06/15 01:46:15  robertj
 * Added channel pause (aka mute) functions.
 *
 * Revision 1.36  2000/06/09 07:50:12  robertj
 * Changed PSTN line psuedo gateway to use LID Dial function.
 *
 * Revision 1.35  2000/06/07 05:48:06  robertj
 * Added call forwarding.
 *
 * Revision 1.34  2000/05/23 12:57:37  robertj
 * Added ability to change IP Type Of Service code from applications.
 *
 * Revision 1.33  2000/05/10 05:16:00  robertj
 * Added hook flash to user indications. Also allowed selection of the User Indication
 *    method to be used in the H.245 protocol.
 *
 * Revision 1.32  2000/05/10 04:06:19  robertj
 * Added ability to disable an audio codec as well as change its priority.
 *
 * Revision 1.31  2000/05/05 03:45:05  robertj
 * Added ability to set multiple aliases for endpoint.
 * Added "bandwidth by type" so can specify by saying "T1" for example.
 *
 * Revision 1.30  2000/05/04 11:53:38  robertj
 * Added Packets Too Late statistics.
 * Fixed gatekeeper detection so if "required", disables call menus.
 *
 * Revision 1.29  2000/04/11 03:59:33  robertj
 * Added new call end reasons for gatekeeper denied calls.
 *
 * Revision 1.28  2000/04/05 20:56:47  robertj
 * Added more statistics.
 *
 * Revision 1.27  2000/03/30 19:55:08  robertj
 * Fixed some call end message strings, also added cap exchange failure message.
 *
 * Revision 1.26  2000/03/04 12:16:55  robertj
 * Added ring sound file to standard sound device on incoming calls.
 *
 * Revision 1.25  2000/02/29 12:58:27  robertj
 * Added extra tracing options to dialog.
 *
 * Revision 1.24  2000/02/17 12:25:17  robertj
 * Added user indication send/receive and option to disable H.245 tunneling.
 *
 * Revision 1.23  2000/02/16 05:21:39  robertj
 * Added dialog parametes for vide transmission, still no grabber yet.
 *
 * Revision 1.22  1999/11/22 10:07:23  robertj
 * Fixed some errors in correct termination states.
 *
 * Revision 1.21  1999/11/22 00:56:35  robertj
 * Improved reason display for connection failure.
 *
 * Revision 1.20  1999/11/19 09:14:53  robertj
 * Added some times for call start and duration.
 *
 * Revision 1.19  1999/11/17 03:49:51  robertj
 * Added RTP statistics display.
 *
 * Revision 1.18  1999/11/16 13:22:24  robertj
 * Improved versioning in about dialog (version number now in single place)
 *
 * Revision 1.17  1999/11/16 11:32:06  robertj
 * Added some calling tones.
 *
 * Revision 1.16  1999/11/13 14:12:24  robertj
 * Changes to make silence detection selectable and disable quicknet option if no cards.
 *
 * Revision 1.15  1999/11/11 23:14:40  robertj
 * Fixed turning off ring on answering call and disabling call menu if initialisation failed.
 *
 * Revision 1.14  1999/11/11 08:54:27  robertj
 * Changed preferred codec function to be full codec ordering.
 *
 * Revision 1.13  1999/11/11 01:20:48  robertj
 * Numerous enhancements to get nearly feature complete for version 1.
 *
 * Revision 1.12  1999/11/06 03:48:36  robertj
 * Added shells for mute functions
 *
 * Revision 1.11  1999/11/05 08:54:41  robertj
 * Rewrite of ixj interface code to fix support for arbitrary codecs.
 *
 * Revision 1.10  1999/11/01 00:50:16  robertj
 * Added receive H.261 video window
 *
 * Revision 1.9  1999/10/30 13:44:04  robertj
 * Fixed duplicate ID in resource strings.
 *
 * Revision 1.8  1999/10/30 12:45:47  robertj
 * Added stop channel message, outside line message and G728/G729 interlock for LineJACK
 *
 * Revision 1.7  1999/10/29 12:08:38  robertj
 * Added bitmap for OpenH323 logo.
 *
 * Revision 1.6  1999/10/29 02:26:52  robertj
 * Added scrollable status window and better termination information.
 *
 * Revision 1.5  1999/10/28 12:21:34  robertj
 * Added AEC support and speakerphone switching button.
 *
 * Revision 1.4  1999/10/24 12:59:40  robertj
 * Added platform independent support for Quicknet xJACK cards.
 *
 * Revision 1.3  1999/10/07 07:37:31  robertj
 * Added auto-answer and trace options (thanks David Iodice of NEWLINK)
 *
 * Revision 1.2  1999/08/27 13:15:43  craigs
 * Removed staticbox and icon from about box as Motif does not yet support them
 *
 * Revision 1.1  1999/08/25 05:16:11  robertj
 * GUI test application.
 *
 * Revision 1.2  1999/01/26 06:35:08  robertj
 * Fixed $LOG$ variable in template files
 *
 */

#include <pwlib/stdres.h>

#inline hdr
#include "main.h"
#inline end


String @IDS_TITLE              "OpenPhone";
String @IDS_LISTENING          "Waiting for incoming call ...";
String @IDS_NOPORT             "Could not open H.323 listener port on %i!";
String @IDS_MAKECALL           "Make Call";
String @IDS_ANSWERCALL         "Answer";
String @IDS_REFUSECALL         "Refuse";
String @IDS_HANGUP             "Hang Up";
String @IDS_SPEAKERPHONE       "Speakerphone";
String @IDS_TRANSFER           "Transfer";
String @IDS_HOLD               "Hold";
String @IDS_INTRUDECALL        "Intrude Call";
String @IDS_RETRIEVE           "Retrieve";
String @IDS_SPEAKER            "Spkr";
String @IDS_MICROPHONE         "Mic";
String @IDS_RINGING            "%s is ringing ...";
String @IDS_TALKING            "Talking to %s";
String @IDS_CALLING            "Calling %s ...";
String @IDS_INTRUDING          "Intruding %s ...";
String @IDS_TRANSFERRING       "Transferring %s ...";
String @IDS_HOSTUNREACHABLE    "%s could not be reached.";
String @IDS_HOSTOFFLINE        "%s is not online.";
String @IDS_NOENDPOINT         "No phone running for %s.";
String @IDS_TRANSPORTERROR     "Transport error calling %s.";
String @IDS_REFUSED            "%s did not accept your call.";
String @IDS_BUSY               "%s was busy.";
String @IDS_CONGESTION         "Congested link to %s.";
String @IDS_NOANSWER           "%s did not answer your call.";
String @IDS_REMOTEHUNGUP       "%s has cleared the call, duration %s.";
String @IDS_REMOTESTOPPED      "%s has stopped calling.";
String @IDS_ABNORMAL           "Call with %s ended abnormally, duration %s";
String @IDS_CAPEXCHFAIL	       "Could not find common codec with %s.";
String @IDS_HUNGUP             "Call with %s completed, duration %s.";
String @IDS_FORWARDING         "Call being forwarded to %s at %s.";
String @IDS_INCOMINGCALL       "Incoming call from %s at %s ..."; 
String @IDS_INCOMINGTRANSFER   "Incoming transferred call from %s at %s ..."; 
String @IDS_IGNORINGCALL       "Incoming call from %s ignored."; 
String @IDS_DECLINED           "Refused incoming call from %s.";
String @IDS_NOUSER             "Could not find user %s.";
String @IDS_NOBANDWIDTH	       "Call to %s refused, no bandwidth available.";
String @IDS_HANGINGUP          "Hanging up connection.";
String @IDS_TRACE_FAIL         "Warning: could not open trace output file \"%s\".";
String @IDS_NOGATEKEEPER       "Could not find gatekeeper!";
String @IDS_INVALIDADDRESS     "Remote address \"%s\" is not valid!";
String @IDS_LIDOPENFAIL        "Could not open %s \"%s\"!";
String @IDS_STARTCHANNELTX     "Started sending %s data%s.";
String @IDS_STARTCHANNELRX     "Started receiving %s data%s.";
String @IDS_STOPCHANNELTX      "Stopped sending %s data.";
String @IDS_STOPCHANNELRX      "Stopped receiving %s data.";
String @IDS_REPLACEHANDSET     "Replace phone handset ...";
String @IDS_ONSPEAKER	       "Switched to speakerphone.";
String @IDS_OUTSIDELINE        "0\tOutside line";
String @IDS_USINGOUTSIDELINE   "Using outside line, waiting for hangup ...";
String @IDS_LINERINGING	       "Outside line ringing ...";
String @IDS_CALLERID           "Outside line caller %s is ringing ...";
String @IDS_LINECALLDIALLED    "Call dialled on local line, please pick up ...";
String @IDS_CALLDISAPPEARED    "Outside line caller stopped ringing.";
String @IDS_USERINPUT          "User input \"%s\" received.";
String @IDS_USINGGK            "Registered with gatekeeper \"%s\".";
String @IDS_UNREGGK	       "Unregistering from gatekeeper \"%s\".";
String @IDS_SENDINGTONE        "Sending tone \"%c\".";
String @IDS_NODIALTONE         "No dial tone!";
String @IDS_LINECALLFAILED     "Error using PSTN line.";
String @IDS_LINEBUSY           "%s is busy.";
String @IDS_INCALLEXIT         "Currently in a call, hang up and exit?";
String @IDS_OPENEDLID          "Opened %s - %s";


MenuBar
{
    Class MainMenu;
    Window MainWindow;

    Menu "Call" {
        Item {
            Title        "Make Call...";
            Notify       "MakeCall";
	    Accelerators "F2";
	    Field        makeCallItem;
        }
        Separator;
        Item {
            Title        "Accept call";
            Notify       "AcceptCall";
	    Accelerators "F3";
        }
        Item {
            Title        "Refuse call";
            Notify       "RefuseCall";
	    Accelerators "F4";
        }
        Separator;
        Menu "Send Indication" {
            Item { Title "1"; Accelerators { "KP1", "1" } Field ind1; Notify UserInputCmd; }
            Item { Title "2"; Accelerators { "KP2", "2" } Field ind2; Notify UserInputCmd; }
            Item { Title "3"; Accelerators { "KP3", "3" } Field ind3; Notify UserInputCmd; }
            Item { Title "4"; Accelerators { "KP4", "4" } Field ind4; Notify UserInputCmd; }
            Item { Title "5"; Accelerators { "KP5", "5" } Field ind5; Notify UserInputCmd; }
            Item { Title "6"; Accelerators { "KP6", "6" } Field ind6; Notify UserInputCmd; }
            Item { Title "7"; Accelerators { "KP7", "7" } Field ind7; Notify UserInputCmd; }
            Item { Title "8"; Accelerators { "KP8", "8" } Field ind8; Notify UserInputCmd; }
            Item { Title "9"; Accelerators { "KP9", "9" } Field ind9; Notify UserInputCmd; }
            Item { Title "0"; Accelerators { "KP0", "0" } Field ind0; Notify UserInputCmd; }
            Item { Title "*"; Accelerators   "KP*";       Field inds; Notify UserInputCmd; }
            Item { Title "#"; Accelerators   "Ret";       Field indh; Notify UserInputCmd; }
            Item { Title "!"; Accelerators   "KP+";       Field indf; Notify UserInputCmd; }
        }
        Item {
            Title        "Show statistics";
            Notify       "ShowStats";
	    Accelerators "F5";
        }
        Separator;
        Item {
            Title        "Mute Microphone";
            Notify       "MuteMike";
	    Accelerators "F7";
        }
        Item {
            Title        "Mute Speaker";
            Notify       "MuteSpeaker";
	    Accelerators "F8";
        }
        Separator;
        Item {
            Title        "Record media...";
            Notify       "RecordMedia";
        }
        Item {
            Title        "Stop recording";
            Notify       "StopRecording";
        }
        Separator;
        Item {
            Title        PSTD_STR_EXIT_MENU;
            Notify       "ExitCmd";
	    Accelerators PSTD_STR_EXIT_ACCEL;
        }
    }

    Menu "Options" {
        Item {
            Title  "Speed Dial...";
            Notify "SpeedDialOptions";
        }
        Separator;
        Item {
            Title  "General...";
            Notify "GeneralOptions";
        }
        Item {
            Title  "Networking...";
            Notify "NetworkingOptions";
        }
        Item {
            Title  "Forwarding...";
            Notify "ForwardingOptions";
        }
        Item {
            Title  "Gatekeeper...";
            Notify "GatekeeperOptions";
        }
        Item {
            Title   "Audio Device...";
            Notify  "AudioDeviceOptions";
        }
        Item {
            Title   "Audio Codec...";
            Notify  "AudioCodecOptions";
        }
        Item {
            Title   "Video...";
            Notify  "VideoOptions";
        }
#ifdef PTRACING
        Separator;
        Item {
            Title   "Trace options...";
            Notify  "TraceOptions";
        }
#endif
    }
}


Icon @IDI_MAIN_WINDOW
{
    Dimensions 64,64,4;
    Colours {
        255, 255, 255,
        132, 140, 181,
         74,  82, 132,
        206, 206, 214,
        239, 239, 255,
        214, 214, 239,
        189, 189, 214,
        140, 140, 165,
        189, 189, 222,
        165, 165, 198,
        107, 107, 132,
        132, 132, 189,
        107, 107, 156,
        132, 132, 198,
         49,  49, 115,
          0,   0,   0
    }
    Pixels {
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  8 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15  5  3 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9  5 15 15 15 15 15 15 15 15 15  6  6  4 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15  6  6  6 15 15 15 15 15 15 15 15 13 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 13 13  5 15 15 15 15 15 15 15  6  6  6  5 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15  6  6  6  6 15 15 15 15 15 15 15 11 13 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13 11  5 15 15 15 15 15 15  6  6  8  6  8 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15  5  6  6  6  6 15 15 15 15 15 15  9 11 13 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 13  7  4 15 15 15 15  4  6  6  6  8  6  4 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15  5  6  6  6  6  6 15 15 15 15 15  1 11 11 11 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13 11 11  1  5 15 15 15  3  6  6  6  8  6  6  4 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15  5  6  6  6  6  8  8  4 15 15 15 11 11 11 11 11 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 13 11 11  1  3 15 15  6  6  6  6  6  6  6  6 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15  6  6  6  6  6  8  6  3 15 15 11 11 11 11 11 13 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13 11 13 13 11  1  6 15  6  6  8  6  8  6  6  8  5 15 15 15 15 15 15
        15 15 15 15 15 15 15  5  6  6  6  6  6  8  6  6 15 11 11 11 13 11 11 11 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 13 11 11 11 11 11 13  1  3  6  6  6  6  8  6  8  8  6  4 15 15 15 15 15
        15 15 15 15 15 15 15  6  6  6  6  8  6  8  8  6  9 13 13 11 11 11 13 13 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 13 13 13 11 11 13  1  6  6  8  6  8  6  6  6  6  6 15 15 15 15 15
        15 15 15 15 15 15  6  6  6  6  6  8  6  8  6  9 13 13 11 11 13 11 11 11 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13 11 11 11 11 13 13 11  1  6  6  6  8  6  6  6  6  6  3 15 15 15 15
        15 15 15 15 15 15  6  6  6  6  6  6  8  6  8 13 11 13 13 13 13 13 11 11 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  7 13 13 13 13 13 13 13 11 11 13  8  6  6  6  8  6  6  6  6  6 15 15 15 15
        15 15 15 15 15  8  6  8  8  8  8  6  6  6  8 13 13 13 13 13 13 13 11 13 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 11 11 13 13 13 11 11 11  9  6  8  6  6  6  8  6  6  6  5 15 15 15
        15 15 15 15 15  6  8  6  8  8  8  6  8  8  9 13 13 11 11 11 11 11 13 13 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 11 13 11 11 11 11 11 11 11  8  6  6  8  6  6  6  6  6  6 15 15 15
        15 15 15 15  8  6  6  8  6  6  6  6  6  8 11 11 13 13 13 11 11 11 11  1  3 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4  7  1 11 11 11 13 13 13 11 11  9  6  8  6  6  8  6  6  6  6  5 15 15
        15 15 15 15  6  8  8  6  8  8  6  6  6  5 11 11 11 11 11 11 11 11  5 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 11 11 11 11 11 13 11 13 13  8  6  6  6  6  6  6  6  8  6 15 15
        15 15 15  4  6  6  6  6  8  8  8  6  4  7 11 13 13 13 13 13  7  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  1 13 13 11 11 11 13 11  4  6  6  6  8  6  6  8  8  6 15 15
        15 15 15  8  6  6  8  6  8  8  6 15 15 11 13 11 11 11 11  7 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  8 11 13 11 11 11 13 13 15  6  6  8  6  6  6  6  6  5 15
        15 15 15  6  8  8  6  6  8  6  8 15  5 11 11 11 11 13 11  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 13 11 11 13 13 11 11 15  4  8  6  6  6  6  6  8  6 15
        15 15 15  6  8  8  6  6  8  6 15 15  9 11 13 13 13 13  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 11 13 11 13 13 13 15 15  4  6  6  6  6  6  6  6 15
        15 15  4  6  6  6  6  6  8 15 15 15  1 11 11 11 11  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 11 11 13 11 13  5 15 15  3  6  6  6  6  6  6 15
        15 15  5  6  6  6  6  6  5 15 15  4 13 13 11 13  9 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4 13 11 11 11 13 15 15 15  6  6  6  8  6  8  5
        15 15  3  6  8  6  6  8 15 15 15  3 11 13 13 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13 13 11 15 15 15  6  6  6  8  6  6  5
        15 15  8  6  6  6  6  6 15 15 15  9 13 11 11  3 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4 11 11 13 11 15 15 15  4  8  6  8  6  8  3
        15 15  8  6  8  6  8  6 15 15 15  7 11 13 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 11 11 13 11  4 15 15 15  6  6  8  6  6  3
        15 15  8  8 11 12 12  3 15 15 15  7 11 11 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  7 13 13 13  4 15 15 15  6  8  6  6  6  6
        15 15  2 14  9  6  6  2  8 15 15  1 11 11 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4  4 15  4 15 15  4  4  4 15 15 15 15 15  7 11 11 11  5 15 15 15  3  8  6  8  8  6
        15  2 14  6  8  8  6 15 14  9 15  7 13 13  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  3 14 12 15 15 15 15 14 10 15 15 15 15  4  5 11 11 11  8  4  4 15  4  6  8  8  9  6
         8 14  9  6  8  6  6 15  2 14  4  1 13 13  6 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 14  9 15 15 15 15 14  3 15 15  5 14 14 14 13 13 11 14 14 14  9  4  6  1 14 14 14
        14 14  5  6  6  6  6 15  3 14 12  1 13  2  9 12 14  8 15 15 15 15 12 14 12 15 15 15  8  5  4  2 12 15 15 15 15 14  9 15 15 15 15 14  3 15 15  7 15 15 14 11 11  2  8 15 15 14  4  8  1  8  6 14
        14 14  4  6  8  6  6 15 15 14 14 12  2 14  1 10 14 14  9 15 15  2 15 15  2 14 15  6 14  9 10  9 14  2 15 15 15 14  9 15 15 15 15 14  3 15 15 15 15 15 14 13 13 12  8 15 15 14  4  6  8  8  8  1
        14 14  0  6  6  6  6  4 15 14 14 11  1 14 12 15 15 14 14 15  5 14 15 15  5 14  8 15  2  3 15 15 10 14 15 15 15 14 11  5  5  5  5 14  3 15 15 15 15 15  5 11 11 11  8 15 15 14  3  6  6  8  8 10
        14 14  0  6  6  6  6  5 15 14 14  1  1 14 11 15 15  2 14 15 11 14 10 10 10 14  1 15 14  5 15 15 12 14 15 15 15 14 14 14 14 14 14 14  3 15 15 15 15  7 12 13 11 11  5 15 15  2  5  6  6  6  2  2
        14 14  0  5  8  6  8  6  4 14  2  1 11 14  1 15 15  9 14 15  2  2 15 15 15 15 15 15 14  3 15 15 12 14 15 15 15 14  9 15 15 15 15 14  3 15 15 15 15  5 14 12 11 13  4 15  6  0  6  8  8  8  9 14
         2 14  4  4  6  6  6  6  6 14  8  7  1 14 11 15 15  6 14 15  2 14 15 15 15 15 15 15 14  5 15 15  7 14 15 15 15 14  9 15 15 15 15 14  3 15 15 15 15 15  2  2 13 11 15 15 10  5  6  8  6  6  4  5
         4 14 10  0  6  8  6  6 14 14 15  6  1 14 11 15 15 13 14 15 13 14 15 15 15 15  9 15 14  5 15 15  1 14 15 15 15 14  9 15 15 15 15 14  3 15 15 15 15 15 10  2 13 11 15  2  0  6  6  6  6  6  0  0
        15  6 14  5  4  8  6  6 14  4 15  5  1 14 12  3 15 14  9 15 15 14 14  6  6 14 15 15 14  3 15 15 12 14 15 15 15 14  9 15 15 15 15 14  3 15 15  9 15 15 14 13 13 11 14  3  6 12 12  6  1  3  0  9
        15 15  5 14  2 12  2 14  8  6 15 15 11 14 14 10 14  7 15 15 15  4 14 14 14  4 15  6 14  2  5  3 14 14  7 15 10 14 14  3 15 15  9 14 14  3 15 14 14 14 11 11 11 14 14 14 14 14  6  8 14 14 14  6
        15 15 15 15 15  4  8  6  6  8  6  4 12 14  1 13  9 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4 11 13 11 11  4  0  6  6  6  8  6  3 15 15 15
        15 15 15 15 15 15  6  6  6  6  8  6  1 14 11 11  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 11 11 11  0  6  6  6  6  6  6  4 15 15 15
        15 15 15 15 15 15  5  6  8  6  8  6  9 14  1 13 13  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  5 11 11 13 13 11  8  6  6  6  6  6  6 15 15 15 15
        15 15 15 15 15 15 15  6  6  6  6  6 12 14 12 11 13 11  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  5 11 11 13 11 11  9  6  8  8  6  6  6 15 15 15 15 15
        15 15 15 15 15 15 15 15  6  6  8  8  8  1 13 11 13 11 13 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  5 11 11 11 13 13 11  8  6  6  6  6  6  4 15 15 15 15 15
        15 15 15 15 15 15 15 15  6  6  6  6  6  6 11 13 13 11 11 11 11  6  6  8 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4  4  5  5 13 11 13 13 13 13 11 13  6  6  8  8  8  5 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15  6  6  8  6  8  1 13 11 11 11 13 13 11  9  6  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  6  6  8 11 13 11 11 11 11 11 11  8  6  6  6  6  9 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15  3  6  8  8  6 13 11 13 13 11 11 11 11 11  8 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  6 11 11 13 11 11 13 13 13 13  9  8  6  8  8  6 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15  6  8  6  8 11 11 11 13 11 13 13 13 13  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15  3  1 11 13 11 11 13 11 11 13 11  8  6  8  6  6  5 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15  3  6  8 13 13 13 13 11 13 11 11 11 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 13 13 13 13 13 13 11 11 11 11  8  6  8  6  4 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15  5  8  8 13 13 11 11 13 13 13 11 11 15 15 15 15 15 15 15 15 15 15 15 15 15 15  7 11 11 11 11 11 11 13 11 11  9  8  6  8 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  8  8 13 13 11 11 13 11 13 11  5 15 15 15 15 15 15 15 15 15 15 15 15 15  7 11 13 13 13 13 13 11 11 11  6  6  4 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4  9 11 13 11 11 11 11 11  8 15 15 15 15 15 15 15 15 15 15 15 15  4  1 11 11 11 11 11 13 11 11  8  8 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  8 11 11 13 11 11 11  8 15 15 15 15 15 15 15 15 15 15 15 15  6 11 11 11 13 13 13 13 11  3 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  8 13 13 13 11 11  8  4 15 15 15 15 15 15 15 15 15 15 15  6 11 13 11 11 11 11  1 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  1 13 13 13 11  5  5 15 15 15 15 15 15 15 15 15 15 15  6  9 13 13 11 13  1  5 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4  9 11 11 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 13 13 13  7 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9  1 13 15 15 15 15 15 15 15 15 15 15 15 15 15 15  9 11 13  7 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  7 15 15 15 15 15 15 15 15 15 15 15 15 15 15  7  1  5 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15  4 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
        15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
    }
    AndMask {
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1
        1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1
        1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1
        1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1
        1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1
        1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1
        1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1
        1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1
        1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1
        1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1
        1 1 1 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 1
        1 1 1 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 1
        1 1 1 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 1
        1 1 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 1
        1 1 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 0 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 0 1 1 0 0 0 1 1 1 1 1 0 0 0 0 0 1 1 1 0 0 0 0 0 0
        1 0 0 0 0 0 0 1 0 0 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 0 0 1 1 1 1 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0
        0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 1 1 1 0 0 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 1 1 1 0 0 0 0 0 1 1 1 1 0 0 1 1 1 1 0 0 1 1 0 1 1 0 0 0 0 0 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 0 0 1 1 0 1 1 0 0 1 0 0 0 0 0 0 0 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 1 0 0 0 0 0 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 1 1 0 0 1 0 0 1 1 0 0 0 1 0 0 1 1 0 0 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 1 1 0 0 1 0 0 0 0 0 0 0 1 0 0 1 1 0 0 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 1 0 0 1 1 1 1 1 1 0 0 1 1 0 0 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 1 0 0 1 1 1 1 1 1 0 0 1 1 0 0 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 1 0 0 0 0 1 1 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 1 1 0 0 1 0 0 1 1 1 1 0 1 0 0 1 1 0 0 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 1 0 0 0 0 1 0 0 0 0 0 0 0 0 0
        1 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 1 0 0 1 1 0 0 0 0 0 1 1 0 0 1 1 0 0 1 1 1 0 0 1 1 1 1 0 0 1 1 0 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        1 1 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 1 1 1 0 0 0 0 0 1 0 0 0 0 0 0 0 0 1 0 0 0 0 1 1 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1
        1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1
        1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1
        1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
    }
    XorMask {
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0
        0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0
        0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0
        0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0
        0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0
        0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0
        0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0
        0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0
        0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0
        0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0
        0 0 0 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 0
        0 0 0 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 0
        0 0 0 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 0
        0 0 1 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 0
        0 0 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 1 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 1 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 1 0 0 1 1 1 0 0 0 0 0 1 1 1 1 1 0 0 0 1 1 1 1 1 1
        0 1 1 1 1 1 1 0 1 1 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 1 1 0 0 0 0 1 1 1 1 1 1 1 1 0 1 1 1 1 1 1
        1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 0 0 1 1 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1 1 1 0 0 0 0 1 1 1 0 0 0 1 1 1 1 1 0 0 0 0 1 1 0 0 0 0 1 1 0 0 1 0 0 1 1 1 1 1 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1 1 1 1 0 0 1 0 0 1 1 0 1 1 1 1 1 1 1 0 0 0 1 1 0 0 0 0 1 1 0 0 0 0 0 1 1 1 1 1 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 0 0 1 1 0 1 1 0 0 1 1 1 0 1 1 0 0 1 1 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 1 1 1 1 1 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 0 1 1 1 1 1 1 0 0 1 1 0 1 1 1 1 1 1 1 0 1 1 0 0 1 1 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 1 1 1 1 1 1 0 0 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 1 0 1 1 0 0 0 0 0 0 1 1 0 0 1 1 0 0 0 1 1 0 0 0 0 1 1 0 0 0 0 1 1 1 1 1 1 0 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 1 1 0 1 1 0 0 0 0 0 0 1 1 0 0 1 1 0 0 0 1 1 0 0 0 0 1 1 0 0 0 0 0 1 1 1 1 0 0 1 1 1 1 1 1 1 1
        1 1 1 1 1 1 1 1 1 1 0 1 1 1 1 0 0 1 1 0 1 1 0 0 0 0 1 0 1 1 0 0 1 1 0 0 0 1 1 0 0 0 0 1 1 0 0 0 0 0 1 1 1 1 0 1 1 1 1 1 1 1 1 1
        0 1 1 1 1 1 1 1 1 1 0 1 1 1 1 1 0 1 1 0 0 1 1 1 1 1 0 0 1 1 0 0 1 1 0 0 0 1 1 0 0 0 0 1 1 0 0 1 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        0 0 1 1 1 1 1 1 1 1 0 0 1 1 1 1 1 1 0 0 0 1 1 1 1 1 0 1 1 1 1 1 1 1 1 0 1 1 1 1 0 0 1 1 1 1 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
        0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0
        0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0
        0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0
        0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0
        0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
    }
}


IMAGE @IDBM_LOGO
{
    Dimensions 150,120,4;
    Colours {
        255, 255, 255,
        132, 140, 181,
         74,  82, 132,
        206, 206, 214,
        239, 239, 255,
        214, 214, 239,
        189, 189, 214,
        140, 140, 165,
        189, 189, 222,
        165, 165, 198,
        107, 107, 132,
        132, 132, 189,
        107, 107, 156,
        132, 132, 198,
         49,  49, 115,
          0,   0,   0
    }
    Pixels {
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  8  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  5  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  6  6  6  3  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  7  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7  1 11  7  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5 13 13 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 13  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  8  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  8  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9  1  1 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11  1  1  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  8  6  8  6  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  6  6  6  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 11 11 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 13 11 11 13 11 11 11 11  5  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  6  6  6  8  6  6  6  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  8  6  8  6  6  4  0  0  0  0  0  0  0  0  0  0  0  4  9  7 11 11 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 11  7  9  4  0  0  0  0  0  0  0  0  0  4  9  6  6  8  6  8  6  6  8  6  6  3  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  8  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  4 11 13 13 11 13 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7 11 13 11 13 11 11 13 13 13 13  9  4  0  0  0  0  0  0  0  0  5  6  6  6  6  6  6  6  8  6  6  8  6  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  6  6  8  6  8  6  6  3  0  0  0  0  0  0  0  0  0  5  1  1 11 11 13 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 11 13 13 11 11 11 11 11  1  1  5  0  0  0  0  0  0  0  3  6  6  6  6  6  8  6  6  8  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  8  6  8  6  6  6  6  6  6  8  0  0  0  0  0  0  0  0  8  1 11 11 13 11 13 11 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 13 11 13 11 13 11 13 11 13 11 11 11  1  5  0  0  0  0  0  4  8  6  8  6  8  6  6  8  6  6  6  8  6  6  6  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  6  6  6  6  6  8  6  8  6  6  4  0  0  0  0  0  0  6 11 11 11 13 11 13 11 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 13 11 13 11 13  1 11  3  0  0  0  0  4  6  6  6  6  6  6  8  6  6  6  8  6  6  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  8  6  6  8  6  6  6  6  6  5  0  0  0  0  0  8  1 11 11 13 11 11 11 13 11 13 11 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7 11 13 11 13 11 13 11 11 11 13 11 11 13 11  1  5  0  0  0  5  6  6  8  6  8  6  6  6  8  6  6  6  8  6  8  6  6  5  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  6  6  6  6  6  8  6  6  8  6  6  8  6  3  0  0  0  0  6 11 11 11 13 11 13 11 11 13 11 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 11 13 11 11 13 11 13 11 13 11 11 13  1  1  6  0  0  6  6  6  6  6  6  8  6  6  6  8  6  6  6  6  6  8  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  8  6  8  6  8  6  6  8  6  6  8  6  6  6  6  4  0  0  5  1 11 11 13 11 11 11 13 13 11 13 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 13 11 13 11 13 11 11 13 11 13 11 11 11  1  5  4  6  6  8  6  8  6  6  6  8  6  6  6  6  6  8  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  8  6  6  6  6  6  6  6  8  6  6  8  6  6  6  6  6  4  0  5 11 11 11 13 11 11 13 13 11 11 13 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 11 13 11 13 11 11 13 11 13 11 11 13 11  1 11  3  6  6  6  6  6  6  6  8  6  6  8  6  6  6  8  6  8  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  8  6  8  6  6  6  8  6  6  8  6  6  6  5  4  1 11 11 13 11 13 11 13 11 13 13 11 11 13 11 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7 11 11 13 11 13 11 11 13 11 13 11 11 13 11 13 11 13 11 13  8  6  8  6  6  8  6  6  6  8  6  6  8  6  6  6  6  6  8  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  8  6  6  6  6  6  6  8  6  6  6  8  6  8  6  6  6  8  9  7 13 11 13 11 11 13 11 13 11 13 11 13 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 11 13 11 13 11 11 13 11 11 13 11  1 13  6  6  8  6  6  8  6  6  6  8  6  6  8  6  8  6  6  6  6  8  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  8  6  6  8  6  8  6  8  6  6  6  6  6  6  8  6  6  8  8  9 11 13 11 13 11 13 11 13 11 13 11 11 13 11 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 13 11 13 11 13 11 11 13 11 13 11 11 13 11 13 13 13 13  8  6  8  6  6  6  8  6  6  8  6  6  6  6  8  6  8  6  6  6  8  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  6  6  6  6  6  8  6  6  6  8  6  6  6  8  9  1 13 11 13 11 11 13 11 11 13 11 11 13 11 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 11 13 11 11 13 11 11 13 11 13 11 11 13 11 13 13 11 13  1  9  8  6  6  6  8  6  6  8  6  6  8  6  8  6  6  6  6  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  6  6  8  6  6  8  6  8  6  8  6  6  8  6  6  6  8  8  9 13 13 11 11 13 13 13 11 13 11 13 11 13 11 13 11 13 11 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7 11 13 11 13 11 13 11 13 13 11 13 11 13 11 13 11 11 11 11 11 13 13  8  8  6  6  6  6  8  6  6  8  6  6  6  6  8  6  6  8  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  8  6  6  8  6  6  6  6  6  6  8  6  6  6  6  6  6  8  9 11 11 13 11 11 13 11 13 11 13 11 11 13 11 11 13 11 13 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 11 13 13 11 11 13 11 11 13 13 11 13 11 13 11  9  6  6  6  8  6  6  8  6  6  8  6  8  6  6  8  6  6  8  6  6  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  6  6  6  8  6  6  8  6  8  6  8  6  6  8  6  8  6  8  8  9 11 13 11 13 11 13 11 13 11 13 11 13 11 13 11 13 11 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 13 11 13 11 13 11 13 11 11 11 13 11 13 11 13 11 11 13 11 11 11 13 11  9  8  6  6  8  6  6  6  6  6  6  6  8  6  6  6  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  8  6  6  6  6  6  6  6  6  6  8  6  6  6  6  6  6 13 11 11 11 11 11 13 11 11 13 11 13 13 11 11 13 11 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 11 13 11 13 11 11 13 11 13 11 11 13 11 11 13 11 11 13 13 11 11 13  8  6  8  6  6  8  6  6  8  6  8  6  6  8  6  8  6  8  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  6  8  6  6  8  6  8  6  8  6  8  6  6  6  8  6  8  8  9 13 11 13 11 13 13 11 13 11 13 11 11 11 13 11 13 11 13 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9  1 11 13 11 13 11 13 11 13 11 11 13 11 13 11 13 11 13 11 13 11 11 13 11  6  8  6  6  8  6  6  8  6  6  6  6  8  6  6  6  6  6  6  8  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  6  6  6  6  6  6  6  8  8  6  6  6  6  8 11 11 11 13 11 11 13 11 11 13 11 13 13 11 11 13 11 11 13 11 11 13 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  9 13 11 11 13 13 11 13 11 13 11 11 13 11 11 13 11 11 13 11 13 11 11 13  1  8  8  6  6  6  6  6  8  6  8  6  6  8  6  8  6  8  6  6  6  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  6  6  8  6  6  8  6  8  6  8  6  8  6  6  6  8  6  8  8  9 11 13 13 11 13 13 11 13 11 13 11 11 11 11 13 11 11 13 11  1  1  7  3  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  7  1  1 11 11 11 13 11 13 11 11 13 11 13 11 13 11 11 13 11 11 13  9  6  6  8  8  8  6  6  6  6  8  6  6  6  6  6  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  6  8  6  6  6  6  6  6  6  6  6  8  6  6  6  6 13 11 11 11 13 11 11 11 13 11 11 13 11 13 13 11 11 13 11 13  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  5 11 13 11 13 11 11 13 11 11 13 11 11 13 11 11 13 11 11 13 11 13  8  6  8  6  6  6  8  6  8  6  6  8  6  6  6  8  6  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  8  6  8  6  6  8  6  8  6  8  6  6  8  6  6  6  8  5 11 11 13 11 11 13 11 13 11 13 11 13 11 11 11 11 11  1  7  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  7 11 11 11 13 11 13 11 13 11 13 13 13 11 13 11 13 11 13  3  8  6  6  8  6  6  6  6  6  8  6  6  8  6  6  6  8  8  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  6  6  6  8  6  6  6  6  6  6  6  6  6  8  6  5  4  9 11 11 11 13 11 11 13 11 11 13 11 11 13 11 13 11 11  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  8 11 11 13 11 11 13 11 13 11 11 13 11 11 13 11 13 13  9  4  8  6  6  8  6  8  6  6  6  8  6  6  8  6  6  6  8  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  6  6  8  6  8  6  8  6  8  6  6  3  4  4  7 11 11 13 11 13 11 13 11 13 11 13 11 13 11  7  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  1  7 13 11 13 11 11 13 11 13 11 13 11 13 11 11  7  4  4  6  6  6  6  6  8  8  6  6  6  6  6  6  8  6  8  6  6  8  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  5  9  6  6  8  6  6  6  8  6  6  6  6  6  6  6  8  3  4  0  8  1 13 13 11 11 13 11 11 13 11 11 13 11 11 13  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6 11 13 11 11 13 11 13 11 13 11 11 13 11 11  1  5  0  4  8  6  8  8  6  6  6  8  6  8  6  8  6  6  6  6  8  6  5  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  8  6  6  6  6  6  8  6  6  6  8  8  6  8  6  6  6  0  0  0 13 11 11 13 11 13 11 13 11 13 11 11 11 11  7  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  7 11 11 13 11 11 13 11 13 13 11 11 13 11 13  0  0  4  6  6  6  6  8  6  6  6  6  6  6  6  8  6  8  6  6  5  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  4  6  6  8  6  8  6  6  8  6  6  6  8  6  6  6  8  4  0  0  5 11 11 13 11 13 11 11 13 11 11 13 11 11  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5 13 11 11 13 11 11 13 11 13 13 11 11 11 11  4  0  0  4  8  8  6  6  8  6  6  8  6  8  6  6  6  6  8  6  6  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  6  8  6  8  6  8  6  6  6  6  6  4  0  0  0  9 11 13 11 13 13 11 11 13 11 13 11 13  7  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  7 11 13 11 13 11 11 11 11 13 11 13 11  6  0  0  0  4  9  6  6  6  8  6  6  6  6  8  6  8  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  3  6  6  8  6  8  6  6  6  6  6  6  8  6  6  5  0  0  0  0  9 13 11 13 13 11 11 13 11 13 11 13  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4 11 11 13 11 11 13 11 13 11 13 11 13  7  0  0  0  0  4  6  6  8  6  6  8  6  8  6  6  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  6  8  6  6  6  6  6  8  6  8  6  8  6  9  5  0  0  0  0  4  7 11 13 11 11 13 13 11 11 11 11  7  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  7 11 13 11 13 11 11 13 11 13 11  7  4  0  0  0  0  5  9  6  8  6  6  6  6  6  8  6  8  6  8  6  3  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  4  6  6  8  6  8  6  6  6  6  6  6  6  8  6  0  0  0  0  0  8  1 13 11 13 11 11 13 11 13 11 13  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 11 13 11 13 11 13 11 11 11 13  7  5  0  0  0  0  0  3  8  6  6  8  6  8  6  6  6  6  6  6  6  8  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  6  8  6  8  6  8  6  6  6  4  0  0  0  0  0  9  1 11 13 11 13 11 13 11 13  7  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  7 13 11 11 13 11 13 11 13 11  9  0  0  0  0  0  4  6  6  6  6  6  6  8  6  8  8  6  6  6  8  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  5  6  6  8  6  8  6  6  6  6  6  6  6  5  0  0  0  0  0  4 13 13 11 13 11 11 13 11 13 11  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  1 13 11 13 11 13 11 13 11 11 13  0  0  0  0  0  0  5  6  6  8  6  8  6  6  8  6  6  6  8  6  5  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  5  6  6  6  8  6  6  8  6  8  6  6  6  4  0  0  0  0  0  5 11 11 13 11 13 11 13 11  1 11  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  1 11 11 11 13 11 13 11 11 13  4  0  0  0  0  0  0  6  6  6  6  6  6  8  6  6  8  6  6  6  5  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  3  6  6  6  8  8  6  6  6  6  8  6  5  0  0  0  0  0  0  3 11 11 11 13 11 13 11 13 11  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 11 13 11 13 11 11  3  0  0  0  0  0  0  5  6  8  6  8  6  6  8  6  6  8  6  6  5  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  8  6  8  6  6  6  8  6  8  6  6  6  4  0  0  0  0  0  0  6 13 11 13 11 11 13 11 11  7  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  7 13 11 13 11 13 11 13 11  6  0  0  0  0  0  0  0  8  9  6  6  8  6  6  8  6  6  6  6  5  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  8  6  6  6  6  8  6  6  6  8  6  3  0  0  0  0  0  0  0  9 11 13 11 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9 11 11 13 11 11 13 11 11  9  0  0  0  0  0  0  0  4  6  8  6  6  6  8  6  6  6  8  6  3  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  8  6  6  8  6  6  6  8  6  6  6  4  0  0  0  0  0  0  4  7 11 11 13 11 13 11 11 11  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5 11 11 13 11 13 11 13 11  9  0  0  0  0  0  0  0  4  6  6  6  8  6  8  6  8  6  6  6  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  8  6  6  6  8  6  6  6  8  6  6  4  0  0  0  0  0  0  4  7 13 11 11 13 11 13  1  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 13 11 13 11 11 13 11 11  7  4  0  0  0  0  0  0  0  3  6  6  6  6  8  6  6  8  6  8  3  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  8  8  8  9 11 12 12 12 12 13  3  0  0  0  0  0  0  0  4  7 13 11 13 11 11 11  1  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  7 11 13 11 13 11 13  7  4  0  0  0  0  0  0  0  4  6  6  8  6  6  8  6  6  6  6  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  6  7  2 14 14  2  2  2 14 14  2 12  5  0  0  0  0  0  5  7 11 13 11 11 13 11  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  1 13 11 13 11 11 13  7  4  0  0  0  0  0  0  0  4  8  6  6  6  8  6  6  6  8  6  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  4  2 14 14 12  9  6  6  8  6  7  2 14  2  8  0  0  0  0  5  1 13 11 13 11 13 11  9  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  4  4  4  4  4  0  4  4  0  0  0  0  4  4  4  4  4  4  4  4  0  0  0  0  0  0  0  0  0  0  4  7 11 11 13 11 13 11  1  5  0  0  0  0  0  0  0  0  3  6  8  6  6  6  8  6  8  6  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  5  2 14  2  9  8  6  6  6  6  6  8  4 10 14 14  9  0  0  0  8  1 11 11 13 11 11 11  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  7  2 14  2 14 14 10  9  3  0  0  0  9  7  2 14  2 14  2 10  7  4  0  0  0  0  0  0  0  0  0  6 11 13 11 11 13 11  1  8  0  0  0  0  0  0  0  0  5  6  6  8  6  6  6  6  6  8  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  5  2 14 14  1  6  8  8  6  8  6  6  8  0  4  2 14 14  9  0  0  8  7 11 13 11 13 11  1  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3 14 14 14 12  4  0  0  0  0  0  0  0  6 14 14 14 10  0  0  0  0  0  0  0  0  4  4  4  0  5 11 11 13 11 13 11 11  8  0  4  4  4  0  0  0  0  4  9  6  6  8  8  8  8  9  8  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0 10 14 14  2  8  8  6  6  6  6  6  6  3  0  0  8 14 14 14  3  0  5  1 13 11 11 11 11 13  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 10 14 14  9  0  0  0  0  0  0  0  0  4  2 14 14  3  0  0  0  0  0  0  5  7 14  2 14 10  9 11 11 13 11 11 13  1  1 10 14  2 14 10  3  0  0  4  6  8  6  8 11  2 14 14 14 10  5  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  8 14 14 14  9  6  6  6  8  6  6  6  6  3  0  0  4  2 14 14 10  4  8  1 11 13 11 13  1  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0  5  2 14 14 14 14 14 14 12 13 11 13 11 11 12 14 14 14 14 14 14 14  9  0  4  9  6  8  1 14 14 14 14 14 14  2  4  0  0  0  0  0  0  0
         0  0  0  0  0  4  2 14 14  2  5  6  6  6  6  8  6  6  6  5  0  0  0  7 14 14 14  8  9 11 11 11 13 11  1  9  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  4  2  7  5  0  5 10 14 14  2 11 13 11 13 12 14  2  9  5  6 10 14 14 14  5  4  6  6  9 14  7  9  8  9  2 14 14  7  0  0  0  0  0  0  0
         0  0  0  0  0  5 14 14 14  1  5  6  6  8  6  6  6  8  6  3  0  0  0  3 14 14 14 12  9  1 13 13 12  2 14  9  0  8 12  2 14 12  8  0  0  0  0  0  0  0  0  4  9 12  2 14  2 12  5  0  0  0  0  0  4  8  1 14  5  0  4  1  2 14 12  8  0  0  0  0  0  0  0  0 10 14 14  9  0  0  0  0  0  0  0  0  4  2 14 14  3  0  0  0  0  6  7  0  0  0  0  0 12 14 14 11 13 11 13  2  2  8  0  0  0  0  7 14 14  1  4  6  8 12  1  8  8  8  6  6 14 14 10  0  0  0  0  0  0  0
         0  0  0  0  0  7 14 14 14  9  4  6  6  6  6  6  8  6  6  5  0  0  0  5 14 14 14  2  8  1  2  2 14 14 14  9  7 14 14 14 14 14 14  1  0  0  0  0  0  0  5 10 14  9  3  9  2 14 14  7  0  0  4  7  2 14 14 14  5  3  2 14 14 14 14 14  6  0  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4 14 14  2  3  0  0  0  0  4  0  0  0  0  0  0  6 14 14 11 11 13 11  2  1  8  0  0  0  0  4 14 14 12  4  6  8  9  6  8  6  6  6  6 12 14 10  0  0  0  0  0  0  0
         0  0  0  0  0 10 14 14 14  6  4  6  6  6  8  6  6  6  6  8  0  0  0  0  2 14 14 14  9 12 12  2 14 14 14  1 14  2 10 10 14 14 14 14  9  0  0  0  0  5  2  2  3  0  0  0  5  2 14 14  9  0  5  6  7 14 14 14  9  2 10  6  9 10 14 14  2  0  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0  0  0  0  0  0  0  8 14  2 13 11 13 11 12  1  8  0  0  0  0  0 10 14 10  4  6  6  6  8  6  8  6  8  8  1 14  6  0  0  0  0  0  0  0
         0  0  0  0  0 10 14 14 14  8  0  3  6  6  6  8  6  8  6  6  4  0  0  4  2 14 14 14 13 11 13  1 14 14 14 14  9  4  0  4  9 14 14 14  2  4  0  0  0 10 14  6  0  0  0  0  0  9 14 14  2  4  0  0  4 14 14 14  2  6  0  0  0  5 14 14 14  5  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4  2 14 14  3  0  0  0  0  0  0  0  0  0  0  0 12  2 11 11 13 11 13 13  1  5  0  0  0  0  0 10 14  1  5  6  6  6  8  6  6  6  6  8  2  7  4  0  0  0  0  0  0  0
         0  0  0  0  0 10 14 14 14  5  0  5  6  8  6  6  6  6  6  6  4  0  0  0  2 14 14 14  1 11 13  1 14 14 14 12  0  0  0  0  0  7 14 14 14  9  0  0  5  2 14  4  0  0  0  0  0  5 14 14 14  8  0  0  5  2 14 14  3  0  0  0  0  0 10 14 14  8  0  0  0  0  0  0 12 14 14 11  5  5  5  5  5  5  5  5  8 14 14 14  3  0  0  0  0  0  0  0  0  0  0  8 10  5 11 11 13 11 11 11  1  8  0  0  0  0  4  2 14  8  3  6  6  8  6  6  8  8  8 12 10  4  0  0  0  0  0  0  0  0
         0  0  0  0  0 10 14 14 14  8  0  5  6  6  6  6  6  8  6  6  5  0  0  0  2 14 14 14  9  1 13  1 14 14 14 11  0  0  0  0  0  4  2 14 14 12  0  0 11 14 14 10 10 10 10 10 10 10 14 14 14  1  0  0  4 14 14 14  5  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  2 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14  3  0  0  0  0  0  0  0  0  4  7 14 14 12 11 13 11 11 13 11  1  5  0  0  0  0  5 14  2  4  5  6  6  8  6  8  6  9  2 14  2  9  0  0  0  0  0  0  0  0
         0  0  0  0  0 10 14 14 14  8  0  4  6  6  8  6  8  6  6  6  6  0  0  4  2 14 14 14  5  1 11 12 14 14 14  1  4  0  0  0  0  0 12 14 14  2  0  0 12 14  2  7  7  7  7  7  7  7  9  9  7  6  0  0  4 14 14 14  5  0  0  0  0  0  1 14 14  9  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  5  0  0  0  0  0  0  0  6  2 14 14 14 14  2 11 11 13 11 11  7  4  4  0  0  0  7 14  3  4  6  6  6  6  6  6 12 14 14 14 14 14  7  0  0  0  0  0  0  0
         0  0  0  0  0  7 14 14 14  9  0  0  5  9  8  6  6  8  8  6  6  4  0  4 14 14 14  2  5  1 13 11 14 14 14  1  3  0  0  0  0  0  9 14 14  2  0  0  2 14  2  0  0  0  0  0  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4  2 14 14  3  0  0  0  0  0  0  0  0  0  5  7 14 14 14 12 13 11 13 13  7  4  0  0  0  6 14  7  0  5  6  6  8  6  8  6  8  8  9 12 14 14 14  4  0  0  0  0  0  0
         0  0  0  0  0  3 14 14 14 12  0  0  5  6  6  8  6  6  6  6  6  5  0  5 14 14 14 12  4  7 13  1 14 14 14 11  9  0  0  0  0  0  6 14 14  2  0  0  2 14  2  4  0  0  0  0  0  0  0  0  0  0  0  0  5  2 14 14  5  0  0  0  0  0  1 14 14  9  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0  0  0  0  0  0  0  9 14 14  2 13 11 11 13  7  4  0  0  4  2 10  0  0  6  6  6  6  6  6  6  8  6  5  0  7 14 14  9  0  0  0  0  0  0
         0  0  0  0  0  4  2 14 14  2  4  0  4  6  6  6  6  8  6  6  6  6  4  6 14 14 14  8  0  7 13  1 14 14 14 11  7  0  0  0  0  0  6 14 14  2  0  0  2 14 14  8  0  0  0  0  0  0  0  0  0  0  0  0  4 14 14 14  5  0  0  0  0  4  7 14 14  9  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4  2 14  2  3  0  0  0  0  0  0  0  0  0  0  0  4  2 14  2 11 13 11 11  9  0  0  0 12 10  4  0  5  6  6  8  8  6  6  8  6  6  4  0  5 14 14 13  0  0  0  0  0  0
         0  0  0  0  0  0  7 14 14 14  5  0  0  3  6  8  6  6  6  8  6  6  5 12 14 14 10  0  4  9 11 12 14 14 14 11  1  4  0  0  0  0  8 14 14  7  0  0 10 14 14  1  0  0  0  0  0  0  0  0  5  9  0  0  4 14 14 14  3  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0  0  0  0  0  0  0  8  2 14  2 13 11 13 11  9  0  0  1 10  4  0  0  8  6  6  6  6  6  8  6  6  6  4  0  0 10 14  9  0  0  0  0  0  0
         0  0  0  0  0  0  4  2 14 14 10  0  0  5  6  6  8  6  6  6  6  6  9 14 14 14  6  0  0  6 13  1 14 14 14 11  1  8  0  0  0  0 13 14 14  9  0  0 13 14 14 14  6  0  0  0  0  0  0  0  7  9  0  0  4 14 14 14  5  0  0  0  0  0  1 14 14  9  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  0  0  0  0  0  0  0 13 10 14  2 13 13 11 11  6  0  9  2  3  0  0  5  6  6  6  8  6  6  6  8  6  5  0  0  0 10 14  6  0  0  0  0  0  0
         0  0  0  0  0  0  0  3 14 14 14  8  0  0  8  9  6  8  6  6  8  8 12 14 14  7  0  0  0  3 11  1 14 14 14 11 11 13  4  0  0  0 12 14  2  4  0  0  5 14 14 14 14  8  0  0  0  0  0  6  2  4  0  0  5  2 14 14  5  0  0  0  0  4  7 14 14  9  0  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  0  0  4  2 14 14  5  0  0  0  0  0  0  0  0  0  0  5 13  2 14  1 11 11 13 11  5  9  2  5  0  0  4  6  8  6 12  8  8  6  8  6  6  4  0  0  4  2  2  4  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  6  2 14  2  5  0  4  6  8  6  6  6  6 12 14 14 10  4  0  0  0  5 11  1 14 14 14 12 11 11  3  0  0  3 14 14  9  0  0  0  0 10 14 14 14 14 10  6  3  6 10 14  7  0  0  0  4 14 14 14  3  0  0  0  0  0 12 14 14  9  0  0  0  0  0  0  2 14 14  9  0  0  0  0  0  0  0  0  4 14 14 14  3  0  0  0  0  5  9  4  0  0  0  9  1 14  2 13 11 13 11 11  9 14  9  3  3  6  9  9 12  2 12  8  6 11  1  6  3  0  0  0  9 14  9  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  9  2 14  2  8  4  8  8  8  8 13  2 14 14 12  6  5  0  0  0  0 11 12 14 14 14  2 11 13  1  4  4  2 14 10  0  0  0  0  0  4  2 14 14 14 14 14 14 14 14 10  4  0  0  0  5 14 14 14  5  0  0  0  0  4 12 14 14 11  0  0  0  0  0  5 14 14 14 11  0  0  0  0  0  0  0  0  5 14 14 14  7  0  0  0  0 10 14  2  9  4  5 12 14  2 11 13 11 11 13 12 14 14 14 14 14 14 14 14 14 14  9  6  8 14 14  2  9  4  4  9  2  7  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  5 10 14 14  2 10 12 12  2 14 14  2  9  8  6  6  3  0  0  0  9 11 14 14 14 14 14  2 10 12 14 14  7  4  0  0  0  0  0  0  4 10 14 14 14 14 14 14 10  4  0  0  4  6 10 14 14 14  2  9  5  0  3  7 14 14 14 14  7  5  0  5  9 10 14 14 14 14  7  3  5  0  0  0  3  9 10 14 14 14 14  7  3  4  0  7 14 14 14 14 14  2 12 11 13 11 13 11 12 14  2 14 14 14 14 14 14 14 14  2  6  6  8  2 14 14 14 14 14 10  6  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  4  9 10 10  2  2 10 12  9  6  6  8  6  6  8  3  4  0  8  7 14 14 14  1 12  2  2  2  7  5  0  0  0  0  0  0  0  0  0  0  6  7 10 10 10  6  4  0  0  0  4  3  3  3  3  3  3  3  3  4  3  3  3  3  3  3  3  5  0  3  3  3  3  3  3  3  3  3  4  0  0  0  3  3  3  3  3  3  3  3  3  4  0  0  4  3  3  9  1 11 11 13 11 13 11 11  1  5  4  0  0  5  6  6  6  6  6  6  6  8  6  6  9  6  3  3  4  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  8  8  8  6  6  8  6  6  8  6  6  3  4  4 12 14 14 14  1 13 13 13 11  9  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  7 11 11 11 13 11 11 13 11  1  4  0  0  3  6  6  6  8  8  6  6  8  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  6  6  8  6  6  8  6  6  8  6  6  5  1 14 14 14 11 13 11 11 13  1  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 13 11  7  0  4  6  6  6  6  6  6  6  6  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  6  8  6  6  6  8  8  1 14 14 14 11 11 13 11 11 11 11  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6 11 11 11 13 11 13 11 13 11 13  6  5  6  6  6  8  6  8  6  8  6  6  8  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  8  6  6  6  8  6  6  6  9 14 14 14  1 13 11 13 13 13 11  1  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5 11 11 13 11 11 11 13 11 13 11 11  6  8  6  6  6  6  6  6  6  6  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  8  6  6  6  8  6  6  8  8  9 14 14 14 11 11 13 11 13 11 11  1 11  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5 11  1 13 11 11 11 13 11 13 11 11 13  8  8  6  8  6  8  6  8  6  8  6  8  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  8  6  6  6  8  6  8 12 14 14 14 12 11 13 11 11 13 13 11 11  1  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  1 11 13 11 13 13 13 11 11 13 11  1  9  8  6  6  8  6  8  6  6  6  6  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  8  6  8  6  6  6 12 12  2  2  2  2  2  2 12 13 11 11 11 13 11  7  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3 11  1 11 11 13 11 11 13 13 11 11 13  1  8  6  8  6  6  6  6  8  6  8  8  6  6  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  6  6  6  6  6  8  6  8  6  8  8  1 13 13 13 13 11 13 13 11 11 13 13 11 11  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  5 11 11 13 11 13 11 11 13 11 11 13 11 11 11  8  6  6  8  6  8  6  6  6  6  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  8  6  8  6  6  6  6  6  8  6 11 11 13 11 13 11 11 11 13 11 11 13  1  1  9  5  4  4  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9  1 11 11 11 13 11 11 13 11 11 13 11 11 13  6  8  8  8  6  6  6  6  6  6  8  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  6  6  8  6  8  6  6  6 11 11 11 11 13 11 13 13 11 13 11 13 11 13 11 13  6  6  6  6  6  8  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  4  4  5  5  5  5  5  8 13 11 11 11 13 11 13 11 13 11 13 11 13 11 11 13  6  6  6  6  6  8  6  8  6  8  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  8  6  6  6  6  6  8  6  8  9 13 11 13 11 13 11 13 11 11 13 11 11 11 11 11  1  9  6  6  6  6  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  9  6  6  6  6  6 13  1 11 11 13 11 11 13 11 11 13 11 11 13 11 13 11  6  6  8  6  8  6  6  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  6  6  6  8  6  6  6  8  8  1 11 11 13 11 11 13 11 13 11 11 13 13 13 11 11 13 13  9  8  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  6  6  8  8  9 11 11 11 13 11 11 13 11 13 11 13 11 13 11 13 11 11  1  8  6  6  6  6  8  6  8  6  6  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  6  8  6  6  6  8  6  8  9 13 13 11 11 13 11 13 11 11 13 11 11 11 11 13 11 11 13 11  9  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  9  1 11 11 11 13 11 13 11 13 11 11 13 11 11 13 11 11 11 13  9  8  6  8  6  6  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  9  6  6  8  6  8  8  6 13 11 13 13 11 11 13 11 13 11 11 13 11 13 11 11 13 11 13 11 11  8  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8  6 13 11 13 11 11 11 13 11 11 13 11 11 13 11 13 11 13 11 13 13 11  9  8  8  6  6  6  8  6  8  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  8  6  6  6  8  8 11 11 13 11 13 11 11 13 11 11 11 13 11 13 11 11 13 11 13  1  1  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3 11  1 11 11 11 13 11 13 11 13 11 11 13 11 11 13 11 11 13 11 11 11  1  8  6  6  6  8  6  6  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  9  6  8  6  6  6  9 11 11 11 13 11 13 11 13 13 13 11 13 11 13 11 13 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 13 11 13 11 11 13 11 13 11 13 11 13 11 11 11 13 11 11  9  8  6  8  6  6  8  8  9  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  3  8  6  6  8  8  6 13 11 13 11 13 11 13 11 11 11 13 13 11 13 11 11 13 11 13 11  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 13 11 13 11 11 13 11 13 11 13 11 11 13 11 11 11 11 11 13  6  8  8  6  6  8  6  6  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  6  6  6  8  9  1 11 11 11 13 11 13 11 13 11 11 11 13 11 13 11 13 11 11  9  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  7 11 11 13 11 11 13 11 13 11 11 13 11 11 13 13 11 13 13 13 11  1  6  8  6  6  8  6  6  6  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  8  8  6  8  8  9 13 11 13 11 11 13 11 11 13 11 13 11 11 13 11 11 13 11  9  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  7 13 11 13 11 13 11 11 13 11 13 11 13 11 11 13 13 11 11 11 13  9  8  6  8  6  6  6  8  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  9  8  6  9 11 11 13 11 13 11 13 11 11 11 13 11 13 11 13 11 11  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9 11 11 11 13 13 11 13 11 11 13 11 11 13 13 11 11 13 11  1 13  6  6  8  6  6  6  3  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  8  6  8 11 13 11 13 11 11 13 11 11 13 11 13 11 11 13 11 11  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  7 11 11 13 13 11 13 11 11 13 11 13 11 13 11 11 13 11 13 11  8  6  6  8  6  8  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  3  6  9  1 11 11 13 11 13 11 13 11 11 13 11 13 11 11 13  9  8  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 11 13 13 11 11 11 11 13 11 13 11 11 13 11 13 11 11 11 11  6  8  6  6  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  5  9 13 11 11 13 11 11 13 11 13 11 11 13 11 11 11  9  8  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  8  1 11 11 13 11 13 11 11 13 11 13 11 11 13 11 11 11 11  9  8  6  8  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6 11  1 13 11 13 11 11 13 11 13 11 13 13 11  9  6  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  8  9 13 11 13 11 13 11 13 11 11 13 11 13 11 13 11 11  9  6  3  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 11 11 13 11 11 13 11 13 11 13 11 11 11  9  8  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  8 11 11 11 13 11 11 13 11 11 13 11 13 11 13 11 11  9  3  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  6  1  1 11 13 11 11 13 11 11 13 11 13  9  8  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  8 11 11 13 11 13 11 13 11 13 11 11 13 11  1 11  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 11 13 11 13 11 11 13 11 11 11 11  9  8  6  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  6  8 11 11 13 11 11 13 11 11 13 11 13 11 13  1  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  1  1 13 11 11 13 11 13 11 11  6  5  3  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  6  8  9 11 13 11 13 11 11 13 11 13 11  1  7  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9 11 13 11 13 11 13 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  3  5  1 11 11 11 13 11 11 11 13 11 11 13  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  4  9  7 11 11 13 11 11 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 13 11 13 11 11  7  6  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  8 13 11 13 11 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 11 13 11 11 13 11 13  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9  1  1 13 13  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  9 13 11 11 13  7  7  9  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  5  9 11 11  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9 11 13 11 11 13  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  3  7  3  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  7  1  1  9  5  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  9  9  5  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  4  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
    }
}




Dialog {
    Title "Gatekeeper Discovery";
    Class DiscoverGatekeeperDlg;
    Dim   160,24;

    CentreText { Pos 0,8; Dim 160,8; Title "Searching for Gatekeeper, please wait..."; }
}


Dialog {
    Title "Line Interface Device";
    Class InitialiseDeviceDlg;
    Dim   180,24;

    CentreText { Pos 0,8; Dim 180,8; Title "Initialising drivers, please wait..."; }
}


ModalDialog {
    Class AboutDialog;
    Title "About OpenPhone";
    Dim   165, 135;

    StaticIcon { Pos 10,10;              Icon IDI_MAIN_WINDOW; }

    CentreText { Pos 30,10;  Dim 131, 8; Title "OpenPhone"; }
    CentreText { Pos 30,20;  Dim 125, 8; Title "Version "; Field version; }

    CentreText { Pos  5,45;  Dim 150,10; Title "This software is free to all personal users."; }
    CentreText { Pos  5,80;  Dim 150, 9; Title "Copyright  1999 Equivalence Pty. Ltd."; }
    CentreText { Pos  5,90;  Dim 150, 9; Title "support@equival.com.au"; }

    PushButton { Pos 55,110; Dim  45,15; Title "Ok";  Options ok, default; }
}


ModalDialog {
    Class MakeCallDialog;
    Title "Make Call";

    RightText  { Pos 10,12; Dim 39,  8; Title "Address:"; }
    ComboBox   { Pos 50,10; Dim 100,48; Value address; Field addressBox; Notify AddressChanged; }
    RightText  { Pos 10,28; Dim 39,  8; Title "Gateway:"; }
    ComboBox   { Pos 50,26; Dim 100,48; Value gateway; Field gatewayBox; }

    PushButton { Pos 20,50; Dim 45, 15; Title "Ok";     Options ok, default; }
    PushButton { Pos 85,50; Dim 45, 15; Title "Cancel"; Options cancel;  }

#inline hdr
  public:
    PIntArray recentCalls;
#inline end
}


ModalDialog {
    Class TransferDialog;
    Title "Transfer";

    RightText  { Pos 10,12; Dim 39,  8; Title "Address:"; }
    ComboBox   { Pos 50,10; Dim 100,48; Value address; Field addressBox; Notify AddressChanged; }
    RightText  { Pos 10,28; Dim 39,  8; Title "Gateway:"; }
    ComboBox   { Pos 50,26; Dim 100,48; Value gateway; Field gatewayBox; }
    RadioButton{ Pos 30,40; Dim 40,12;  Title "H.450";  Value useH450; }
    RadioButton{ Pos 70,40; Dim 50,12;  Title "FACILITY"; }

    PushButton { Pos 20,60; Dim 45, 15; Title "Ok";     Options ok, default; }
    PushButton { Pos 85,60; Dim 45, 15; Title "Cancel"; Options cancel;  }

#inline hdr
  public:
    PIntArray recentCalls;
#inline end
}


ModalDialog {
    Class IntrudeCallDialog;
    Title "Intrude Call";

    RightText  { Pos 10,12;  Dim 39,  8; Title "Address:"; }
    ComboBox   { Pos 50,10;  Dim 100,48; Value address; Field addressBox; Notify AddressChanged; }
    RightText  { Pos 10,28;  Dim 39,  8; Title "Gateway:"; }
    ComboBox   { Pos 50,26;  Dim 100,48; Value gateway; Field gatewayBox; }

    RightText   { Pos 5,44;  Dim 59, 8; Title "Capability Level:"; }
    RadioButton { Pos 70,42; Dim 29,12; Title "Low"; Value capabilityLevel; }
    RadioButton { Pos 100,42;Dim 29,12; Title "Med"; }
    RadioButton { Pos 130,42;Dim 35,12; Title "High"; }

    RightText   { Pos 5,65;  Dim 59, 8; Title "Intrusion type:"; }
    RadioButton { Pos 70,60; Dim 80,12; Title "Forced Release"; Value intrusionType; }
    RadioButton { Pos 70,70; Dim 80,12; Title "Held type"; }

    PushButton { Pos 20,90;  Dim 45, 15; Title "Ok";     Options ok, default; }
    PushButton { Pos 85,90;  Dim 45, 15; Title "Cancel"; Options cancel;  }

#inline hdr
  public:
    PIntArray recentCalls;
#inline end
}

ModalDialog {
    Class SpeedDialOptionDlg;
    Title "Speed Dial";
    Dim 212,180;

    RightText  { Pos 10,102;  Dim 39,  8; Title "Alias:"; }
    EditBox    { Pos 50,100;  Dim 60, 12; Value alias; }
    RightText  { Pos 10,117;  Dim 39,  8; Title "Address:"; }
    EditBox    { Pos 50,115;  Dim 120,12; Value address; Notify AddressChanged; }
    RightText  { Pos 10,132;  Dim 39,  8; Title "Gateway:"; }
    EditBox    { Pos 50,130;  Dim 120,12; Value gateway; }

    // Put list box here so alias edit box is first in tab order
    ListBox    { Pos  10,10;  Dim 202,50; Value selectedSpeedDial; Field speedDialList; Notify SpeedDialSelect; }
    PushButton { Pos  10,80;  Dim 42, 14; Title "Add";       Field addButton;    Notify AddButton; Options default; }
    PushButton { Pos  60,80;  Dim 42, 14; Title "Remove";    Field removeButton; Notify RemoveButton; }
    PushButton { Pos 110,80;  Dim 42, 14; Title "Move Up";   Field upButton;     Notify UpButton; }
    PushButton { Pos 160,80;  Dim 42, 14; Title "Move Down"; Field downButton;   Notify DownButton; }

    PushButton { Pos  40,150; Dim 45, 15; Title "Ok";     Options ok; }
    PushButton { Pos 125,150; Dim 45, 15; Title "Cancel"; Options cancel;  }

#inline hdr
  public:
    void EnableFields();
#inline end
}


Dialog {
    Class SpeedDialSelectDlg;
    Title "Speed Dial";
    Dim 200,150;

    LeftText  { Pos  10,10; Dim 180,8;   Title "Enter speed dial number followed by the # key."; }
    RightText { Pos 160,10; Dim 30,8;    Field enteredDTMF; }
    ListBox   { Pos  10,20; Dim 180,120; Field speedDialList; Options MultiColumn; }

#inline hdr
    void OnCancel();
#inline end
}


ModalDialog {
    Class GeneralOptionsDlg;
    Title "General Options";

    RightText  { Pos 10,12;  Dim 59, 8; Title "Username:"; }
    EditBox    { Pos 70,10;  Dim 60,12; Value username; Notify ChangedUsername; }

    RightText  { Pos 10,25;  Dim 59, 8; Title "Aliases:"; }
    EditBox    { Pos 70,25;  Dim 60,12; Value alias; Notify ChangedAlias; }
    PushButton { Pos 70,40;  Dim 30,13; Title "Add"; Notify AddAlias; Field addAliasBtn; }
    PushButton { Pos 100,40; Dim 30,13; Title "Del"; Notify DelAlias; Field delAliasBtn; }
    ListBox    { Pos 140,25; Dim 75,30; Field aliasList; Notify SelectAlias; }

    RightText  { Pos 10,65;  Dim 59, 8; Title "Max. Recent Calls:"; }
    IntEdit    { Pos 70,63;  Dim 30,12; Value maxRecentCalls; Limits 1,100,1,10; }

    RightText  { Pos 10,82;  Dim 59, 8; Title "Ring Sound File:"; }
    EditBox    { Pos 70,80;  Dim 123,12;Value ringSoundFile; Notify ChangedSoundFile; }
    PushButton { Pos 85,95;  Dim 39,13; Title "Browse"; Notify BrowseSound; }
    PushButton { Pos 139,95; Dim 39,13; Title "Play";   Notify PlaySoundFile; Field playSoundButton; }

    CheckBox   { Pos 20,113; Dim 99,12; Title "Auto-Answer";        Value autoAnswer; }
    CheckBox   { Pos 20,124; Dim 99,12; Title "Disable Fast-Start"; Value noFastStart; }
    CheckBox   { Pos 20,135; Dim 99,12; Title "Disable H.245 Tunneling"; Value noTunneling; }
    CheckBox   { Pos 20,146; Dim 99,12; Title "Disable H.245 in SETUP"; Value noH245inSetup; }

    RadioButton{ Pos 125,113; Dim 89,12; Title "DTMF as Q.931 Keypad IE"; Value userInputMode; }
    RadioButton{ Pos 125,124; Dim 89,12; Title "DTMF as H.245 String";}
    RadioButton{ Pos 125,135; Dim 89,12; Title "DTMF as H.245 Signal";}
    RadioButton{ Pos 125,146; Dim 89,12; Title "DTMF as RFC2833";}

    LeftText    { Pos 10,160; Dim 130,8; Title "Call Intrusion Protection Level:"; }
    RadioButton { Pos 53,169; Dim 32,12; Title "Low"; Value intrusionProtectionLevel; }
    RadioButton { Pos 86,169; Dim 32,12; Title "Med"; }
    RadioButton { Pos 123,169;Dim 36,12; Title "High"; }
    RadioButton { Pos 160,169;Dim 36,12; Title "Full"; }

    PushButton { Pos 45,185; Dim 45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 130,185;Dim 45,15; Title "Cancel"; Options cancel;  }
}


ModalDialog {
    Class NetworkingOptionsDlg;
    Title "Networking Options";

    RightText  { Pos 10,12;  Dim 59, 8; Title "Bandwidth:"; }
    ChoiceBox  { Pos 70,10;  Dim 60,60; Value bandwidthType; Notify BandwidthChange;
                 StringList { "14.4k", "28.8k/56k",
                              "64k ISDN", "128k ISDN",
                              "T1/DSL", "LAN", "Custom"
                            }
               }
    FloatEdit  { Pos 140,10; Dim 30,12; Value bandwidth; Field bandwidthBox; Limits 1,99999,10,10; }
    LeftText   { Pos 172,12; Dim 30, 8; Title "kb/s"; }

    RightText  { Pos 10,37;  Dim 59, 8; Title "TCP port base:"; }
    IntEdit    { Pos 70,35;  Dim 26,12; Value tcpBase; Limits 0,65500,1,10; }
    RightText  { Pos 100,37; Dim 54, 8; Title "TCP port max:"; }
    IntEdit    { Pos 155,35; Dim 26,12; Value tcpMax; Limits 0,65500,1,10; }

    RightText  { Pos 10,57;  Dim 59, 8; Title "UDP port base:"; }
    IntEdit    { Pos 70,55;  Dim 26,12; Value udpBase; Limits 0,65500,1,10; }
    RightText  { Pos 100,57; Dim 54, 8; Title "UDP port max:"; }
    IntEdit    { Pos 155,55; Dim 26,12; Value udpMax; Limits 0,65500,1,10; }

    RightText  { Pos 10,77;  Dim 59, 8; Title "RTP port base:"; }
    IntEdit    { Pos 70,75;  Dim 26,12; Value rtpBase; Limits 1024,65500,1,10; }
    RightText  { Pos 100,77; Dim 54, 8; Title "RTP port max:"; }
    IntEdit    { Pos 155,75; Dim 26,12; Value rtpMax; Limits 1024,65500,1,10; }

    RightText  { Pos 10,97;  Dim 94, 8; Title "IP Type Of Service (TOS):"; }
    IntEdit    { Pos 105,95; Dim 18,12; Value ipTOS; Limits 0,255,1,16; }
    LeftText   { Pos 125,97; Dim 30, 8; Title "(hex)"; }

    PushButton { Pos 45,125; Dim 45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 130,125;Dim 45,15; Title "Cancel"; Options cancel;  }
}


ModalDialog {
    Title "Forwarding Options";
    Class ForwardingOptionsDlg;

    RightText  { Pos 10,12; Dim 59, 8; Title "Always:"; }
    EditBox    { Pos 70,10; Dim 80,12; Value alwaysForwardParty; }
    RightText  { Pos 10,27; Dim 59, 8; Title "On Busy:"; }
    EditBox    { Pos 70,25; Dim 80,12; Value busyForwardParty; }
    RightText  { Pos 10,42; Dim 59, 8; Title "No Answer:"; }
    EditBox    { Pos 70,40; Dim 80,12; Value noAnswerForwardParty; }
    RightText  { Pos 10,57; Dim 59, 8; Title "No Answer:"; }
    IntEdit    { Pos 70,55; Dim 18,12; Value noAnswerTime; Limits 0,999,1,10; }
    LeftText   { Pos 90,57; Dim 30, 8; Title "seconds"; }

    PushButton { Pos 25,80; Dim 45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 85,80; Dim 45,15; Title "Cancel"; Options cancel;  }
}


ModalDialog {
    Title "Gatekeeper Options";
    Class GatekeeperOptionsDlg;

    CheckBox   { Pos 20,10; Dim 70,12; Title "Use Gatekeeper"; Value useGatekeeper; Notify UsingGatekeeper; }
    CheckBox   { Pos 90,10; Dim 80,12; Title "Require Gatekeeper"; Value requireGatekeeper; Field requireBox; }
    RadioButton{ Pos 10,26; Dim 90,12; Title "Discover Automatically"; Value discover; Field discoverButton; Notify CheckDiscover; }
    RadioButton{ Pos 10,43; Dim 69, 8; Title "Static Host:"; Field staticGkButton; Notify CheckDiscover; }
    RadioButton{ Pos 10,60; Dim 69, 8; Title "Locate by ID:"; Field idButton; Notify CheckDiscover; }
    EditBox    { Pos 80,41; Dim 85,12; Value gatekeeperHost; Field gatekeeperHostBox; }
    EditBox    { Pos 80,58; Dim 85,12; Value gatekeeperId; Field gatekeeperIdBox; }
    RightText  { Pos 10,79; Dim 69, 8; Title "H.235 Password:"; }
    PasswordBox{ Pos 80,77; Dim 60,12; Value password; }
    RightText  { Pos 10,96; Dim 69, 8; Title "Time To Live:"; }
    IntEdit    { Pos 80,94; Dim 18,12; Value timeToLive; Limits 0,3600,1,10; }
    LeftText   { Pos 100,96;Dim 30, 8; Title "seconds"; }
    RightText  { Pos 10,112;Dim 69, 8; Title "Access token OID:"; }
    EditBox    { Pos 80,110;Dim 85,12; Value accessTokenOID; }

    PushButton { Pos 30,132; Dim 45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 100,132;Dim 45,15; Title "Cancel"; Options cancel;  }

#inline hdr
  public:
    void EnableFields();
#inline end
}


ModalDialog {
    Class AudioDeviceOptionsDlg;
    Title "Audio Device Options";

    RadioButton{ Pos  10,10;  Dim  60,12; Title "Sound Card";    Notify SelectCard; Value cardSelection; }
    RadioButton{ Pos  10,62;  Dim  60,12; Title "Quicknet Card"; Notify SelectCard; Field quicknetButton; }
    RadioButton{ Pos  10,98;  Dim  60,12; Title "VOIP Blaster";  Notify SelectCard; Field blasterButton; }

    RightText  { Pos  70,12;  Dim  49, 8; Title "Playback:"; Field playDeviceTitle; }
    ChoiceBox  { Pos 120,10;  Dim 140,60; Value playSoundDevice; Field playDeviceBox;  }
    RightText  { Pos  70,27;  Dim  49, 8; Title "Recording:"; Field recordDeviceTitle; }
    ChoiceBox  { Pos 120,25;  Dim 140,60; Value recordSoundDevice; Field recordDeviceBox;  }
    RightText  { Pos  70,42;  Dim  49, 8; Title "Sound buffers:"; Field soundBuffersTitle; }
    IntEdit    { Pos 120,40;  Dim  20,12; Value soundBuffers; Field soundBuffersBox; Limits 2,99,1,10; }

    RightText  { Pos  70,62;  Dim  49, 8; Title "Identifier:"; Field xJackDeviceTitle; }
    ChoiceBox  { Pos 120,60;  Dim 140,60; Value xJackDevice; Field xJackDeviceBox;  }
    RightText  { Pos   0,78;  Dim 119, 8; Title "Echo Cancellation:"; Field aecTitle; }
    ChoiceBox  { Pos 120,76;  Dim  40,60; Value aecLevel; Field aecBox;
					  StringList { "Off", "Low", "Medium", "High" } }
    RightText  { Pos 160,78;  Dim  34, 8; Title "Country:"; Field countryTitle; }
    ComboBox   { Pos 195,76;  Dim  65,60; Value country; Field countryBox;
			                  StringList { "Australia",
                                                       "France",
                                                       "Germany",
                                                       "Japan",
                                                       "United Kingdom",
                                                       "United States" } }

    RightText  { Pos  70,98;  Dim  49, 8; Title "Name:"; Field blasterDeviceTitle; }
    ChoiceBox  { Pos 120,96;  Dim 140,35; Value blasterDevice; Field blasterDeviceBox;  }

    PushButton { Pos  55,116; Dim  45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 150,116; Dim  45,15; Title "Cancel"; Options cancel;  }
}


ModalDialog {
    Class AudioCodecOptionsDlg;
    Title "Audio Codec Options";

    LeftText   { Pos  20,10;  Dim  69, 8; Title "Codec Selection:"; }
    ListBox    { Pos  20,19;  Dim 120,40; Field codecList; Notify SelectCodec; }
    PushButton { Pos 150,25;  Dim  39,14; Title "Up";   Field up;   Notify MoveUp; }
    PushButton { Pos 195,25;  Dim  39,14; Title "Down"; Field down; Notify MoveDown; }
    CheckBox   { Pos 160,45;  Dim  40,12; Title "&Enabled"; Field enabledCodec; Value codecEnabled; Notify EnableCodec; }
    IntEdit    { Pos 205,45;  Dim  20,12; Field packetSizeBox; Value packetSize; Limits 1,256,1,10; Notify PacketSizeChange; }
    LeftText   { Pos 227,47;  Dim  23, 8; Title "frames"; }

    RightText  { Pos   0,72;  Dim  54, 8; Title "Jitter Buffer:"; }
    IntEdit    { Pos  55,70;  Dim  30,12; Value jitter; Limits 10,1000,10,10; }
    LeftText   { Pos  87,72;  Dim  15, 8; Title "ms"; }

    CheckBox   { Pos 112,65;  Dim  40,20; Title "Silence"; Value silenceDetect; }
    LeftText   { Pos 124,79;  Dim  36, 8; Title "detection"; }
    RightText  { Pos 155,65;  Dim  29, 8; Title "Signal:"; }
    HScrollBar { Pos 185,63;  Dim  50,12; Value signalDeadband;  Limits 1,2000,8,80; Notify SignalDeadbandChange; }
    LeftText   { Pos 236,65;  Dim  24, 8; Title "000 ms"; Field signalDeadbandMS; }
    RightText  { Pos 155,81;  Dim  29, 8; Title "Silence:"; }
    HScrollBar { Pos 185,79;  Dim  50,12; Value silenceDeadband; Limits 1,8000,8,80; Notify SilenceDeadbandChange; }
    LeftText   { Pos 236,81;  Dim  24, 8; Title "000 ms"; Field silenceDeadbandMS; }

    PushButton { Pos  55,100; Dim  45,15; Title "Ok";     Options ok, default; }
    PushButton { Pos 150,100; Dim  45,15; Title "Cancel"; Options cancel;  }
}


ModalDialog {
    Class VideoOptionsDlg;
    Title "Video Options";

    StaticBox   { Pos 10,10;  Dim 90,65; Title "Receive Size"; }
    CheckBox    { Pos 25,25;  Dim 60,12; Title "Small (QCIF)"; Value rxqcif; }
    CheckBox    { Pos 25,40;  Dim 60,12; Title "Large (CIF)";  Value rxcif; }
    CheckBox    { Pos 25,55;  Dim 60,12; Title "Auto Receive";  Value rxauto; }

    StaticBox   { Pos 110,10; Dim 90,65; Title "Transmit Size"; }
    RadioButton { Pos 125,25; Dim 60,12; Title "Disabled"; Value txsize; }
    RadioButton { Pos 125,40; Dim 60,12; Title "Small (QCIF)"; }
    RadioButton { Pos 125,55; Dim 60,12; Title "Large (CIF)"; }

    RightText  { Pos   0,85;  Dim  59, 8; Title "Device:"; }
    ChoiceBox  { Pos  60,83;  Dim 120,60; Value videoDevice; Field videoDeviceBox;  }

    RightText   { Pos 10,106; Dim 29, 8; Title "Format:"; }
    ChoiceBox   { Pos 40,104; Dim 40,60; Value format;  StringList { "PAL", "NTSC", "SECAM", "Auto" } }

    RightText   { Pos 81,106; Dim 28, 8;  Title "Source:"; }
    ChoiceBox   { Pos 110,104;Dim 30,60; Value source;  StringList { "Any", "A", "B", "C" } }

    RightText   { Pos 141,106;Dim 28, 8; Title "Quality:"; }
    IntEdit     { Pos 170,104;Dim 25,12; Value quality; Limits 0,31,1,10; }

    CheckBox    { Pos 20,120;  Dim 50,12; Title "Local Video"; Value localVideo; }
    CheckBox    { Pos 80,120;  Dim 50,12; Title "Flip Local";  Value flipLocal; }
    CheckBox    { Pos 140,120; Dim 50,12; Title "Flip Remote";  Value flipRemote; }

    PushButton  { Pos 40,140; Dim 45,15; Title "Ok";     Options ok, default; }
    PushButton  { Pos 125,140;Dim 45,15; Title "Cancel"; Options cancel;  }
}


#ifdef PTRACING

ModalDialog {
    Class TraceOptionsDlg;
    Title "Trace Options";

    CheckBox   { Pos 30, 10; Dim  60, 8; Title "Enable Tracing"; Value traceFlag; Notify TraceSet; }
    RightText  { Pos 10, 27; Dim  39, 8; Title "Trace Level:";   Field traceLevelText; }
    IntEdit    { Pos 50, 25; Dim  20,12; Value traceLevel;       Field traceLevelBox;  Limits 0,10,1,10; }
    RightText  { Pos 10, 42; Dim  39, 8; Title "Output File:";   Field traceOutText;   }
    EditBox    { Pos 50, 40; Dim 150,12; Value traceFile;        Field traceOutBox;    }
    CheckBox   { Pos 30, 60; Dim  70,10; Title "Blocks";         Field traceBlocksBox; Value traceBlocks; }
    CheckBox   { Pos 30, 73; Dim  70,10; Title "Date && Time";   Field dateAndTimeBox; Value dateAndTime; }
    CheckBox   { Pos 110,73; Dim  70,10; Title "Timestamp";      Field timestampBox;   Value timestamp; }
    CheckBox   { Pos 30, 86; Dim  70,10; Title "Thread Names";   Field threadsBox;     Value threads; }
    CheckBox   { Pos 110,86; Dim  70,10; Title "Thread Addresss";Field threadAddrBox;  Value threadAddr; }
    CheckBox   { Pos 30, 99; Dim  70,10; Title "Level Number";   Field levelDisplayBox;Value levelDisplay; }
    CheckBox   { Pos 110,99; Dim  70,10; Title "File && Line";   Field fileAndLineBox; Value fileAndLine; }

    PushButton { Pos 45,120; Dim 45, 15; Title "Ok";     Options ok, default; }
    PushButton { Pos 135,120;Dim 45, 15; Title "Cancel"; Options cancel;  }
}

#endif


Dialog {
    Class StatisticsDlg;
    Title "RTP Statistics";

    StaticBox { Pos 10,10; Dim 100,75;Title "Transmitter"; }
    RightText { Pos 20,20; Dim 49,8;  Title "Packets:"; }
    LeftText  { Pos 70,20; Dim 38,8;  Field packetsSent; }
    RightText { Pos 20,32; Dim 49,8;  Title "Octets:"; }
    LeftText  { Pos 70,32; Dim 38,8;  Field octetsSent; }
    RightText { Pos 20,44; Dim 49,8;  Title "Average time:"; }
    CenterText{ Pos 70,44; Dim 24,8;  Field averageSendTime; }
    LeftText  { Pos 95,44; Dim 10,8;  Title "ms"; }
    RightText { Pos 20,56; Dim 49,8;  Title "Maximum time:"; }
    CenterText{ Pos 70,56; Dim 24,8;  Field maximumSendTime; }
    LeftText  { Pos 95,56; Dim 10,8;  Title "ms"; }
    RightText { Pos 20,68; Dim 49,8;  Title "Minimum time:"; }
    CenterText{ Pos 70,68; Dim 24,8;  Field minimumSendTime; }
    LeftText  { Pos 95,68; Dim 10,8;  Title "ms"; }

    StaticBox { Pos 120,10; Dim 110,120; Title "Receiver"; }
    RightText { Pos 130,20; Dim 49,8; Title "Packets:"; }
    LeftText  { Pos 180,20; Dim 38,8; Field packetsReceived; }
    RightText { Pos 130,32; Dim 49,8; Title "Octets:"; }
    LeftText  { Pos 180,32; Dim 38,8; Field octetsReceived; }
    RightText { Pos 130,44; Dim 49,8; Title "Average time:"; }
    CenterText{ Pos 180,44; Dim 24,8; Field averageReceiveTime; }
    LeftText  { Pos 205,44; Dim 10,8; Title "ms"; }
    RightText { Pos 130,56; Dim 49,8; Title "Maximum time:"; }
    CenterText{ Pos 180,56; Dim 24,8; Field maximumReceiveTime; }
    LeftText  { Pos 205,56; Dim 10,8; Title "ms"; }
    RightText { Pos 130,68; Dim 49,8; Title "Minimum time:"; }
    CenterText{ Pos 180,68; Dim 24,8; Field minimumReceiveTime; }
    LeftText  { Pos 205,68; Dim 10,8; Title "ms"; }
    RightText { Pos 130,80; Dim 49,8; Title "Jitter:"; }
    CenterText{ Pos 180,80; Dim 24,8; Field jitter; }
    LeftText  { Pos 205,80; Dim 10,8; Title "ms"; }
    RightText { Pos 130,92; Dim 49,8; Title "Packets lost:"; }
    LeftText  { Pos 180,92; Dim 38,8; Field packetsLost; }
    RightText { Pos 130,104;Dim 49,8; Title "Packets late:"; }
    LeftText  { Pos 180,104;Dim 38,8; Field packetsTooLate; }
    RightText { Pos 130,116;Dim 49,8; Title "Out of order:"; }
    LeftText  { Pos 180,116;Dim 38,8; Field packetsOutOfOrder; }

    RightText { Pos 10,92; Dim 59,8; Title "Round trip time:"; }
    CenterText{ Pos 70,92; Dim 24,8; Field roundTripDelay; }
    LeftText  { Pos 95,92; Dim 10,8; Title "ms"; }

    RightText { Pos 10,105; Dim 59,8; Title "Call duration:"; }
    LeftText  { Pos 70,105; Dim 40,8; Field callDuration; }

#inline hdr
  public:
    void OnCancel();
#inline end
}


///////////////////////////////////////////////////////////////////////////////
